//
// GCALDaemon is an OS-independent Java program that offers two-way
// synchronization between Google Calendar and various iCalalendar (RFC 2445)
// compatible calendar applications (Sunbird, Rainlendar, iCal, Lightning, etc).
//
// http://gcaldaemon.sourceforge.net
//
package org.gcaldaemon.core.ldap;

import java.io.File;
import java.io.FileOutputStream;
import java.io.RandomAccessFile;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.StringTokenizer;

import net.fortuna.ical4j.model.DateTime;
import net.sf.gm.ContactType;
import net.sf.gm.beans.GMContact;

import org.apache.commons.codec.net.QuotedPrintableCodec;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.gcaldaemon.core.Configurator;
import org.gcaldaemon.core.FilterMask;
import org.gcaldaemon.core.GmailEntry;
import org.gcaldaemon.core.GmailPool;
import org.gcaldaemon.core.StringUtils;
import org.gcaldaemon.logger.QuickWriter;

/**
 * Periodic Gmail contact loader thread.
 * 
 * Created: Jan 03, 2007 12:50:56 PM
 * 
 * @author Andras Berkes
 */
public final class ContactLoader extends Thread {

	// --- CONSTANTS ---

	private static final String NATIVE_CHARSET = Charset.defaultCharset()
			.name();
	private static final String VCARD_EXTENSION = ".vcf";

	private static final byte VCARD_UTF8_ENCODING = 0;
	private static final byte VCARD_NATIVE_ENCODING = 1;
	private static final byte VCARD_QUOTED_ENCODING = 2;

	// --- LOGGER ---

	private static final Log log = LogFactory.getLog(ContactLoader.class);

	// --- VARIABLES ---

	private final Configurator configurator;
	private final LDAPListener ldapListener;
	private final String username;
	private final String password;
	private final long pollingTimeout;
	private final File vcardDirectory;
	private final byte vcardEncoding;
	private final String vcardVersion;

	private volatile GMContact[] contacts;

	// --- CONSTRUCTOR ---

	public ContactLoader(ThreadGroup mainGroup, Configurator configurator)
			throws Exception {
		super(mainGroup, "Contact loader");
		this.configurator = configurator;
		this.vcardDirectory = new File(configurator.getWorkDirectory(), "vcard");
		if (!vcardDirectory.isDirectory()) {
			vcardDirectory.mkdirs();
		}

		// Acceptable hostnames
		FilterMask[] hosts = configurator
				.getFilterProperty(Configurator.LDAP_ALLOWED_HOSTNAMES);

		// Acceptable TCP/IP addresses
		FilterMask[] addresses = configurator
				.getFilterProperty(Configurator.LDAP_ALLOWED_ADDRESSES);

		// Get contact list cache timeout
		long timeout = configurator.getConfigProperty(
				Configurator.LDAP_CACHE_TIMEOUT, 3600000L);
		if (timeout < 180000L) {
			log.warn("The fastest contact list polling period is '3 min'!");
			timeout = 180000L;
		}
		pollingTimeout = timeout;

		// Get username
		username = configurator.getConfigProperty(
				Configurator.LDAP_GOOGLE_USERNAME, null);
		if (username == null) {
			throw new NullPointerException("Missing username ("
					+ Configurator.LDAP_GOOGLE_USERNAME + ")!");
		}

		// Get password
		password = configurator
				.getPasswordProperty(Configurator.LDAP_GOOGLE_PASSWORD);

		if (hosts == null && addresses == null) {

			// Security warning
			log.warn("Set the '"
					+ Configurator.LDAP_ALLOWED_HOSTNAMES
					+ "' parameter to limit remote access.");
		} else {

			// Debug filters
			if (log.isDebugEnabled()) {
				log.debug("Allowed LDAP hosts: "
						+ configurator.getConfigProperty(
								Configurator.LDAP_ALLOWED_HOSTNAMES, "*"));
				log.debug("Allowed LDAP addresses: "
						+ configurator.getConfigProperty(
								Configurator.LDAP_ALLOWED_ADDRESSES, "*"));
			}
		}

		// Get vCard properties
		String value = configurator.getConfigProperty(
				Configurator.LDAP_VCARD_ENCODING, "quoted");
		if (value.equals("quoted")) {
			vcardEncoding = VCARD_QUOTED_ENCODING;
		} else {
			if (value.equals("native")) {
				vcardEncoding = VCARD_NATIVE_ENCODING;
			} else {
				vcardEncoding = VCARD_UTF8_ENCODING;
			}
		}
		value = configurator.getConfigProperty(Configurator.LDAP_VCARD_VERSION,
				"3.0");
		try {
			double num = Double.parseDouble(value);
			vcardVersion = Double.toString(num);
		} catch (Exception formatError) {
			log.fatal("Invalid vCard version: " + value);
			throw formatError;
		}

		// Create and start LDAP listener
		int port = (int) configurator.getConfigProperty(Configurator.LDAP_PORT,
				9080);
		ldapListener = new LDAPListener(this, hosts, addresses, port);

		// Start listener
		start();
	}

	// --- CONTACT LOADER LOOP ---

	public final void run() {
		for (;;) {
			try {

				// Load contact list
				for (int tries = 0;; tries++) {
					try {
						loadContacts();
						break;
					} catch (Exception loadError) {
						if (tries == 5) {
							throw loadError;
						}
						log.debug("Connection refused, reconnecting...");
						Thread.sleep(500);
					}
				}

				// Wait
				sleep(pollingTimeout);

			} catch (InterruptedException interrupt) {
				return;
			} catch (Exception loadError) {
				log.error("Unable to load contact list!", loadError);
				try {
					sleep(pollingTimeout);
				} catch (Exception interrupt) {
					return;
				}
			}
		}
	}

	// --- CONTACT LIST LOADER ---

	private final void loadContacts() throws Exception {

		// Loading contact list
		log.debug("Loading Gmail contact list...");
		GmailPool pool = configurator.getGmailPool();
		GmailEntry entry = null;
		Iterator iterator;
		try {
			entry = pool.borrow(username, password);
			iterator = entry.client.getContacts(ContactType.CONTACT_ALL)
					.iterator();
		} finally {
			pool.recycle(entry);
		}
		String rev = new DateTime().toString();
		LinkedList contactList = new LinkedList();
		HashSet cardFiles = new HashSet();
		String note, email;
		GMContact contact;
		while (iterator.hasNext()) {
			contact = (GMContact) iterator.next();
			email = contact.getEmail();
			if (email != null && email.indexOf('@') != -1) {
				contactList.addLast(contact);
			} else {
				note = contact.getNotes();
				if (note != null) {
					String[] emails = findEmailsInNote(note, null);
					if (emails != null) {
						contact.setEmail(emails[0]);
						contactList.addLast(contact);
					} else {

						// Save vcard without email address
						if (!"cle".equals(contact.getId())) {
							cardFiles.add(saveVCard(contact, rev));
						}
					}
				}
			}
		}
		GMContact[] array = new GMContact[contactList.size()];
		contactList.toArray(array);

		// Save contacts withall email addresses
		int i;
		for (i = 0; i < array.length; i++) {
			cardFiles.add(saveVCard(array[i], rev));
		}

		// Remove deleted contacts
		String[] currentFiles = vcardDirectory.list();
		String fileName;
		for (i = 0; i < currentFiles.length; i++) {
			fileName = currentFiles[i];
			if (fileName.endsWith(VCARD_EXTENSION)
					&& !cardFiles.contains(fileName)) {
				(new File(vcardDirectory, fileName)).delete();
			}
		}

		// Contact list loaded
		synchronized (this) {
			contacts = array;
		}
		log.debug(array.length + " contacts loaded successfully.");
	}

	private final String saveVCard(GMContact contact, String rev) {
		String name = contact.getEmail().toLowerCase();
		if (name.length() == 0) {
			name = contact.getName().toLowerCase();
		}
		name = name.trim();
		if (name.length() == 0 || name.indexOf('=') != -1) {
			return VCARD_EXTENSION;
		}
		char[] chars = name.toCharArray();
		QuickWriter writer = new QuickWriter(chars.length);
		boolean writeMinus = true;
		char c;
		int i;
		for (i = 0; i < chars.length; i++) {
			c = chars[i];
			if (c != '_' && Character.isJavaIdentifierPart(c)) {
				writer.write(c);
				writeMinus = true;
				continue;
			}
			if (c == ',') {
				break;
			}
			if (writeMinus) {
				writer.write('-');
				writeMinus = false;
			}
		}
		name = writer.toString() + VCARD_EXTENSION;
		File file = new File(vcardDirectory, name);
		FileOutputStream out = null;
		try {
			writer = new QuickWriter(500);
			String encoding = StringUtils.UTF_8;
			String displayName = contact.getName();
			if (displayName.length() == 0) {
				return VCARD_EXTENSION;
			}
			String email = contact.getEmail();
			String note = contact.getNotes();
			String[] phones = null;
			String[] emails = null;
			if (note != null) {
				note = note.replace('\r', ' ').replace('\n', ' ');
				note = note.trim();
				if (note.length() == 0) {
					note = null;
				} else {
					phones = findPhoneNumbersInNote(note);
					emails = findEmailsInNote(note, email);
				}
			}
			String firstName = null;
			String lastName = null;
			i = displayName.indexOf(' ');
			if (i != -1) {
				firstName = displayName.substring(0, i);
				lastName = displayName.substring(i + 1);
			}

			// Write vCard
			writer.write("BEGIN:VCARD\r\n");
			if (vcardVersion.charAt(0) == '3') {
				writer.write("VERSION:");
				writer.write(vcardVersion);
				writer.write("\r\nPRODID:");
				writer.write(Configurator.VERSION);
				writer.write("\r\n");
			} else {
				writer.write("VERSION:2.1\r\n");
			}
			switch (vcardEncoding) {
			case VCARD_UTF8_ENCODING:

				// Pure UTF8 vCard format
				writer.write("X-LOTUS-CHARSET:UTF-8\r\n");
				writer.write("FN;CHARSET=UTF-8:");
				writer.write(displayName);
				if (firstName != null) {
					writer.write("\r\nN;CHARSET=UTF-8:");
					writer.write(firstName);
					writer.write(';');
					writer.write(lastName);
					writer.write(";;;");
				}
				if (note != null) {
					writer.write("\r\nNOTE;CHARSET=UTF-8:");
					writer.write(note);
				}
				break;
			case VCARD_NATIVE_ENCODING:

				// Native vCard format
				encoding = NATIVE_CHARSET;
				writer.write("X-LOTUS-CHARSET:");
				writer.write(NATIVE_CHARSET);
				writer.write("\r\nFN:");
				writer.write(displayName);
				i = displayName.indexOf(' ');
				if (firstName != null) {
					writer.write("\r\nN:");
					writer.write(firstName);
					writer.write(';');
					writer.write(lastName);
				}
				if (note != null) {
					writer.write("\r\nNOTE:");
					writer.write(note);
				}
				break;
			default:

				// Quoted-printable vCard format
				encoding = StringUtils.US_ASCII;
				writer.write("X-LOTUS-CHARSET:UTF-8\r\n");
				writer.write("FN;QUOTED-PRINTABLE:");
				writer.write(encodeQuotedPrintable(displayName));
				i = displayName.indexOf(' ');
				if (firstName != null) {
					writer.write("\r\nN;QUOTED-PRINTABLE:");
					writer.write(encodeQuotedPrintable(firstName));
					writer.write(';');
					writer.write(encodeQuotedPrintable(lastName));
				}
				if (note != null) {
					writer.write("\r\nNOTE;QUOTED-PRINTABLE:");
					writer.write(encodeQuotedPrintable(note));
				}
			}
			if (email.length() != 0) {

				// Default email
				if (vcardVersion.charAt(0) == '3') {
					writer.write("\r\nEMAIL;TYPE=PREF;TYPE=INTERNET:");
				} else {
					writer.write("\r\nEMAIL;PREF;INTERNET:");
				}
				writer.write(email);
			}
			if (emails != null) {

				// Additional emails
				for (i = 0; i < emails.length; i++) {
					if (vcardVersion.charAt(0) == '3') {
						writer.write("\r\nEMAIL;TYPE=INTERNET:");
					} else {
						writer.write("\r\nEMAIL;INTERNET:");
					}
					writer.write(emails[i]);
				}
			}
			if (phones != null) {
				if (vcardVersion.charAt(0) == '3') {
					writer.write("\r\nTEL;TYPE=HOME:");
				} else {
					writer.write("\r\nTEL;HOME:");
				}
				writer.write(phones[0]);
				if (phones.length > 1) {
					if (vcardVersion.charAt(0) == '3') {
						writer.write("\r\nTEL;TYPE=WORK:");
					} else {
						writer.write("\r\nTEL;WORK:");
					}
					writer.write(phones[1]);
				}
			}
			writer.write("\r\nREV:");
			writer.write(rev);
			writer.write("\r\nEND:VCARD\r\n");
			byte[] bytes;
			if (encoding.equals(StringUtils.US_ASCII)) {
				bytes = writer.getBytes();
			} else {
				bytes = StringUtils.encodeString(writer.toString(), encoding);
			}
			for (int retries = 0;; retries++) {
				try {
					out = new FileOutputStream(file);
					out.write(bytes);
					out.flush();
					out.close();
					out = null;
					break;
				} catch (Exception lockedError) {
					if (out != null) {
						try {
							out.close();
						} catch (Exception ignored) {
						}
						out = null;
					}
					if (retries == 5) {
						throw lockedError;
					}
					Thread.sleep(500);
				}
			}
		} catch (Exception ioError) {
			log.warn(ioError);
			if (file != null) {
				if (out != null) {
					try {
						out.close();
					} catch (Exception ignored) {
					}
				}
				file.delete();
			}
		}
		return name;
	}

	private static final String encodeQuotedPrintable(String string)
			throws Exception {
		byte[] bytes = StringUtils.encodeString(string, StringUtils.UTF_8);
		bytes = QuotedPrintableCodec.encodeQuotedPrintable(null, bytes);
		return StringUtils.decodeToString(bytes, StringUtils.US_ASCII);
	}

	static final String[] findPhoneNumbersInNote(String note) {
		LinkedList list = new LinkedList();
		StringTokenizer st = new StringTokenizer(note, ",");
		QuickWriter writer = new QuickWriter(50);
		char[] chars;
		char c;
		int i;
		while (st.hasMoreTokens()) {
			chars = st.nextToken().toCharArray();
			for (i = 0; i < chars.length; i++) {
				c = chars[i];
				if (c == '+' || Character.isDigit(c)) {
					writer.write(c);
				} else {
					if (c == '.' || c == ':') {
						if (writer.length() > 6) {
							list.addLast(writer.toString());
						}
						writer.flush();
					}
				}
			}
			if (writer.length() > 6) {
				list.addLast(writer.toString());
			}
			writer.flush();
		}
		int len = list.size();
		if (len == 0) {
			return null;
		}
		String[] array = new String[len];
		list.toArray(array);
		return array;
	}

	static final String[] findEmailsInNote(String note, String defaultEmail) {
		LinkedList list = new LinkedList();
		StringTokenizer st = new StringTokenizer(note, ",");
		String token;
		while (st.hasMoreTokens()) {
			token = st.nextToken().trim();
			if (token.length() > 4 && token.indexOf('.') != -1
					&& token.indexOf('@') != -1) {
				if (defaultEmail != null && defaultEmail.equals(token)) {
					continue;
				}
				list.addLast(token);
			}
		}
		int len = list.size();
		if (len == 0) {
			return null;
		}
		String[] array = new String[len];
		list.toArray(array);
		return array;
	}

	// --- STOP SERVICE ---

	public final void interrupt() {

		// Close server socket and stop listener
		if (ldapListener != null) {
			try {
				ldapListener.interrupt();
			} catch (Exception closeError) {
				log.debug(closeError);
			}
		}

		// Interrupt thread
		super.interrupt();
	}

	// --- GMAIL CONTACT GETTER ---

	public final synchronized GMContact[] getContacts() {
		if (contacts == null) {
			try {

				// Network down - load contacts from vcards
				log.debug("Loading contacts from vcard files...");
				String[] vcards = vcardDirectory.list();
				Arrays.sort(vcards, String.CASE_INSENSITIVE_ORDER);
				LinkedList cardList = new LinkedList();
				String fileName, email;
				GMContact contact;
				for (int i = 0; i < vcards.length; i++) {
					fileName = vcards[i];
					if (fileName.endsWith(VCARD_EXTENSION)) {
						contact = loadVCard(fileName);
						if (contact != null) {
							email = contact.getEmail();
							if (email != null && email.indexOf('@') != -1) {
								cardList.addLast(contact);
							}
						}
					}
				}
				contacts = new GMContact[cardList.size()];
				cardList.toArray(contacts);
				log.debug(contacts.length + " vcards loaded successfully.");
			} catch (Exception ioError) {
				log.warn(ioError);
			}
		}
		return contacts;
	}

	private final GMContact loadVCard(String fileName) {
		File file = null;
		RandomAccessFile raf = null;
		try {

			// Load vcard
			file = new File(vcardDirectory, fileName);
			raf = new RandomAccessFile(file, "r");
			byte[] bytes = new byte[(int) raf.length()];
			raf.readFully(bytes);
			raf.close();
			raf = null;

			// Parse vcard file
			byte format = VCARD_NATIVE_ENCODING;
			String vcard = StringUtils.decodeToString(bytes,
					StringUtils.US_ASCII);
			if (vcard.indexOf("QUOTED-PRINTABLE") != -1) {
				format = VCARD_QUOTED_ENCODING;
			} else {
				if (vcard.indexOf("X-LOTUS-CHARSET:UTF-8") != -1) {
					format = VCARD_UTF8_ENCODING;
					vcard = StringUtils
							.decodeToString(bytes, StringUtils.UTF_8);
				} else {
					vcard = StringUtils.decodeToString(bytes, NATIVE_CHARSET);
				}
			}
			StringTokenizer st = new StringTokenizer(vcard, "\r\n");
			String line, name = null;
			String email = null;
			String note = null;
			int i;
			while (st.hasMoreTokens()) {
				line = st.nextToken().trim();
				i = line.indexOf(':');
				if (i == -1) {
					continue;
				}
				if (line.startsWith("FN")) {
					name = line.substring(i + 1);
					if (format == VCARD_QUOTED_ENCODING) {
						name = decodeQuotedPrintable(name);
					}
					continue;
				}
				if (line.startsWith("EMAIL")) {
					email = line.substring(i + 1);
					continue;
				}
				if (line.startsWith("NOTE")) {
					note = line.substring(i + 1);
					if (format == VCARD_QUOTED_ENCODING) {
						note = decodeQuotedPrintable(note);
					}
				}
			}
			if (name != null && email != null && name.length() != 0
					&& email.length() != 0) {
				GMContact contact = new GMContact();
				contact.setName(name);
				contact.setEmail(email);
				if (note != null && note.length() != 0) {
					contact.setNotes(note);
				}
				return contact;
			}
		} catch (Exception ioError) {

			// Damaged vcard file
			log.warn(ioError);
			if (raf != null) {
				try {
					raf.close();
				} catch (Exception ignored) {
				}
			}
			if (file != null) {
				file.delete();
			}
		}
		return null;
	}

	private static final String decodeQuotedPrintable(String string)
			throws Exception {
		byte[] bytes = StringUtils.encodeString(string, StringUtils.US_ASCII);
		bytes = QuotedPrintableCodec.decodeQuotedPrintable(bytes);
		return StringUtils.decodeToString(bytes, StringUtils.UTF_8);
	}

}
